# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from lxml import etree  # pytype: disable=import-error
from typing import (
    Any,
    Container,
    Generator,
    Iterable,
    MutableMapping,
    Optional,
    Tuple,
)
from picosvg.geometric_types import Rect


SVGCommand = Tuple[str, Tuple[float, ...]]
SVGCommandSeq = Iterable[SVGCommand]
SVGCommandGen = Generator[SVGCommand, None, None]


def svgns():
    return "http://www.w3.org/2000/svg"


def xlinkns():
    return "http://www.w3.org/1999/xlink"


def splitns(name):
    qn = etree.QName(name)
    return qn.namespace, qn.localname


def strip_ns(tagname):
    return splitns(tagname)[1]


# https://www.w3.org/TR/SVG11/paths.html#PathData
_CMD_ARGS = {
    "m": 2,
    "z": 0,
    "l": 2,
    "h": 1,
    "v": 1,
    "c": 6,
    "s": 4,
    "q": 4,
    "t": 2,
    "a": 7,
}
_CMD_ARGS.update({k.upper(): v for k, v in _CMD_ARGS.items()})


def check_cmd(cmd, args):
    cmd_args = num_args(cmd)
    if cmd_args == 0:
        if args:
            raise ValueError(f"{cmd} has no args, {len(args)} invalid")
    elif len(args) % cmd_args != 0:
        raise ValueError(f"{cmd} has sets of {cmd_args} args, {len(args)} invalid")
    return cmd_args


def num_args(cmd):
    if not cmd in _CMD_ARGS:
        raise ValueError(f'Invalid svg command "{cmd}"')
    return _CMD_ARGS[cmd]


def cmds():
    return _CMD_ARGS.keys()


# For each command iterable of x-coords and iterable of y-coords
# Helpful if you want to adjust them
_CMD_COORDS = {
    "m": ((0,), (1,)),
    "z": ((), ()),
    "l": ((0,), (1,)),
    "h": ((0,), ()),
    "v": ((), (0,)),
    "c": ((0, 2, 4), (1, 3, 5)),
    "s": ((0, 2), (1, 3)),
    "q": ((0, 2), (1, 3)),
    "t": ((0,), (1,)),
    "a": ((5,), (6,)),
}
_CMD_COORDS.update({k.upper(): v for k, v in _CMD_COORDS.items()})


def cmd_coords(cmd):
    if not cmd in _CMD_ARGS:
        raise ValueError(f'Invalid svg command "{cmd}"')
    return _CMD_COORDS[cmd]


def ntos(n: float) -> str:
    # strip superflous .0 decimals
    return str(int(n)) if isinstance(n, float) and n.is_integer() else str(n)


def number_or_percentage(s: str, scale=1) -> float:
    return float(s[:-1]) / 100 * scale if s.endswith("%") else float(s)


def path_segment(cmd, *args):
    # put commas between coords, spaces otherwise, author readability pref
    args_per_cmd = check_cmd(cmd, args)
    args = [ntos(a) for a in args]
    combined_args = []
    xy_coords = set(zip(*_CMD_COORDS[cmd]))
    if args_per_cmd:
        for n in range(len(args) // args_per_cmd):
            sub_args = args[n * args_per_cmd : (n + 1) * args_per_cmd]
            i = 0
            while i < len(sub_args):
                if (i, i + 1) in xy_coords:
                    combined_args.append(f"{sub_args[i]},{sub_args[i+1]}")
                    i += 2
                else:
                    combined_args.append(sub_args[i])
                    i += 1
    return cmd + " ".join(combined_args)


def parse_css_declarations(
    style: str,
    output: MutableMapping[str, Any],
    property_names: Optional[Container[str]] = None,
) -> str:
    """Parse CSS declaration list into {property: value} dict.

    Args:
        style: CSS declaration list without the enclosing braces,
            as found in an SVG element's "style" attribute.
        output: a dictionary where to store the parsed properties.
        property_names: optional set of property names to limit the declarations
            to be parsed; if not provided, all will be parsed.

    Returns:
        A string containing the unparsed style declarations, if any.

    Raises:
        ValueError if CSS declaration is invalid and can't be parsed.

    References:
    https://www.w3.org/TR/SVG/styling.html#ElementSpecificStyling
    https://www.w3.org/TR/2013/REC-css-style-attr-20131107/#syntax
    """
    unparsed = []
    for declaration in style.split(";"):
        if declaration.count(":") == 1:
            property_name, value = declaration.split(":")
            property_name = property_name.strip()
            if property_names is None or property_name in property_names:
                output[property_name] = value.strip()
            else:
                unparsed.append(declaration.strip())
        elif declaration.strip():
            raise ValueError(f"Invalid CSS declaration syntax: {declaration}")
    return "; ".join(unparsed) + ";" if unparsed else ""


def parse_view_box(s: str) -> Rect:
    box = tuple(float(v) for v in re.split(r",|\s+", s))
    if len(box) != 4:
        raise ValueError(f"Unable to parse viewBox: {s!r}")
    return Rect(*box)


# sentinel object to check if special linked fields such as fx/fy are explicitly set;
# using a float type instead of None to make the typechecker happy, and also so that one
# doesn't need to unwrap Optional type whenever accessing those fields
class _LinkedDefault(float):
    def __new__(cls, attr_name):
        self = float.__new__(cls, "NaN")
        self.attr_name = attr_name
        return self

    def __call__(self, data_obj):
        return getattr(data_obj, self.attr_name)


_ATTRIB_DEFAULTS = {
    "clip-path": "",
    "clip-rule": "nonzero",
    "fill": "black",
    "fill-opacity": 1.0,
    "fill-rule": "nonzero",
    "stroke": "none",
    "stroke-width": 1.0,
    "stroke-linecap": "butt",
    "stroke-linejoin": "miter",
    "stroke-miterlimit": 4,
    "stroke-dasharray": "none",
    "stroke-dashoffset": 0.0,
    "stroke-opacity": 1.0,
    "opacity": 1.0,
    "transform": "",
    "style": "",
    "display": "inline",
    "d": "",
    "id": "",
}


def attrib_default(name: str, default: Any = ()) -> Any:
    if name in _ATTRIB_DEFAULTS:
        return _ATTRIB_DEFAULTS[name]
    if default == ():
        raise ValueError(f"No entry for '{name}' and no default given")
    return default



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Functions meant to help discern if shapes are the same or not.

import copy
import dataclasses
from itertools import islice
from math import atan2
from picosvg.geometric_types import Vector, almost_equal
from picosvg.svg_types import SVGShape, SVGPath
from typing import Generator, Iterable, Optional, Tuple
from picosvg import svg_meta
from picosvg.svg_transform import Affine2D


_SIGNIFICANCE_FACTOR = 5  # Must be at least N x tolerance to be significant
_ROUND_RANGE = range(3, 13)  # range of rounds to try


def _first_move(path: SVGPath) -> Tuple[float, float]:
    cmd, args = next(iter(path))
    if cmd.upper() != "M":
        raise ValueError(f"Path for {path} should start with a move")
    return args


def _vectors(path: SVGPath) -> Generator[Vector, None, None]:
    for cmd, args in path:
        x_coord_idxs, y_coord_idxs = svg_meta.cmd_coords(cmd)
        if cmd.lower() == "z":
            yield Vector(0.0, 0.0)
        else:
            yield Vector(args[x_coord_idxs[-1]], args[y_coord_idxs[-1]])


def _nth_vector(path: SVGPath, n: int) -> Vector:
    return next(islice(_vectors(path), n, n + 1))


def _angle(v: Vector) -> float:
    # gives the directional angle of vector (unlike acos)
    return atan2(v.y, v.x)


def _affine_vec2vec(initial: Vector, target: Vector) -> Affine2D:
    affine = Affine2D.identity()

    # rotate initial to have the same angle as target (may have different magnitude)
    angle = _angle(target) - _angle(initial)
    affine = Affine2D.identity().rotate(angle)
    vec = affine.map_vector(initial)

    # scale to target magnitude
    s = 0
    if vec.norm() != 0:
        s = target.norm() / vec.norm()

    affine = Affine2D.compose_ltr((affine, Affine2D.identity().scale(s, s)))

    return affine


def _first_y(vectors: Iterable[Vector], tolerance: float) -> Optional[Vector]:
    tolerance = _SIGNIFICANCE_FACTOR * tolerance
    for idx, vec in enumerate(vectors):
        if idx > 0 and abs(vec.y) > tolerance:
            return vec
    return None


# Makes a shape safe for a walk with _affine_callback
def _affine_friendly(shape: SVGShape) -> SVGPath:
    path = shape.as_path()
    if shape is path:
        path = copy.deepcopy(path)
    return (
        path.relative(inplace=True)
        .explicit_lines(inplace=True)
        .expand_shorthand(inplace=True)
    )


# Transform all coords in an affine-friendly path
def _affine_callback(affine, subpath_start, curr_pos, cmd, args, *_unused):
    x_coord_idxs, y_coord_idxs = svg_meta.cmd_coords(cmd)
    # hard to do things like rotate if we have 1d coords
    assert len(x_coord_idxs) == len(y_coord_idxs), f"{cmd}, {args}"

    args = list(args)  # we'd like to mutate 'em
    for x_coord_idx, y_coord_idx in zip(x_coord_idxs, y_coord_idxs):
        if cmd == cmd.upper():
            # for an absolute cmd allow translation: map_point
            new_x, new_y = affine.map_point((args[x_coord_idx], args[y_coord_idx]))
        else:
            # for a relative coord no translate: map_vector
            new_x, new_y = affine.map_vector((args[x_coord_idx], args[y_coord_idx]))

        if almost_equal(new_x, 0):
            new_x = 0
        if almost_equal(new_y, 0):
            new_y = 0
        args[x_coord_idx] = new_x
        args[y_coord_idx] = new_y
    return ((cmd, args),)


def normalize(shape: SVGShape, tolerance: float) -> SVGShape:
    """Build a version of shape that will compare == to other shapes even if offset,
    scaled, rotated, etc.

    Intended use is to normalize multiple shapes to identify opportunity for reuse."""

    path = _affine_friendly(dataclasses.replace(shape, id=""))

    # Make path relative, with first coord at 0,0
    x, y = _first_move(path)
    path.move(-x, -y, inplace=True)

    # Normlize vector 1 to [1 0]; eliminates rotation and uniform scaling
    vec1 = _nth_vector(path, 1)  # ignore M 0,0
    affine1 = _affine_vec2vec(vec1, Vector(1, 0))
    path.walk(lambda *args: _affine_callback(affine1, *args))

    # Scale first y movement to 1.0
    vecy = _first_y(_vectors(path), tolerance)
    if vecy and not almost_equal(vecy.y, 1.0):
        affine2 = Affine2D.identity().scale(1, 1 / vecy.y)
        path.walk(lambda *args: _affine_callback(affine2, *args))

    # TODO: what if shapes are the same but different start point
    # TODO: what if shapes are the same but different drawing cmds
    # This DOES happen in Noto; extent unclear

    path.round_multiple(tolerance, inplace=True)
    return path


def _apply_affine(affine: Affine2D, s: SVGPath) -> SVGPath:
    s_prime = copy.deepcopy(s)
    s_prime.walk(lambda *args: _affine_callback(affine, *args))
    return s_prime


def _try_affine(affine: Affine2D, s1: SVGPath, s2: SVGPath, tolerance: float):
    return _apply_affine(affine, s1).almost_equals(s2, tolerance)


def _round(affine, s1, s2, tolerance):
    # TODO bsearch?
    for i in _ROUND_RANGE:
        rounded = affine.round(i)
        if _try_affine(rounded, s1, s2, tolerance):
            return rounded
    return affine  # give up


def affine_between(s1: SVGShape, s2: SVGShape, tolerance: float) -> Optional[Affine2D]:
    """Returns the Affine2D to change s1 into s2 or None if no solution was found.

    Intended use is to call this only when the normalized versions of the shapes
    are the same, in which case finding a solution is typical

    """
    s1 = dataclasses.replace(s1, id="")
    s2 = dataclasses.replace(s2, id="")

    if s1.almost_equals(s2, tolerance):
        return Affine2D.identity()

    s1 = _affine_friendly(s1)
    s2 = _affine_friendly(s2)

    s1x, s1y = _first_move(s1)
    s2x, s2y = _first_move(s2)

    affine = Affine2D.identity().translate(s2x - s1x, s2y - s1y)
    if _try_affine(affine, s1, s2, tolerance):
        return affine

    # Normalize first edge.
    # Fixes rotation, x-scale, and uniform scaling.
    s1_vec1 = _nth_vector(s1, 1)
    s2_vec1 = _nth_vector(s2, 1)

    s1_to_origin = Affine2D.identity().translate(-s1x, -s1y)
    s2_to_origin = Affine2D.identity().translate(-s2x, -s2y)
    s1_vec1_to_s2_vec1 = _affine_vec2vec(s1_vec1, s2_vec1)

    # Move to s2 start
    origin_to_s2 = Affine2D.identity().translate(s2x, s2y)

    affine = Affine2D.compose_ltr((s1_to_origin, s1_vec1_to_s2_vec1, origin_to_s2))
    if _try_affine(affine, s1, s2, tolerance):
        return _round(affine, s1, s2, tolerance)

    # Could be non-uniform scaling and/or mirroring
    # Scale first y movement (after matching up vec1) to match

    # Rotate first edge to lie on x axis
    s2_vec1_angle = _angle(s2_vec1)
    rotate_s2vec1_onto_x = Affine2D.identity().rotate(-s2_vec1_angle)
    rotate_s2vec1_off_x = Affine2D.identity().rotate(s2_vec1_angle)

    affine = Affine2D.compose_ltr(
        (s1_to_origin, s1_vec1_to_s2_vec1, rotate_s2vec1_onto_x)
    )
    s1_prime = _apply_affine(affine, s1)

    affine = Affine2D.compose_ltr((s2_to_origin, rotate_s2vec1_onto_x))
    s2_prime = _apply_affine(affine, s2)

    s1_vecy = _first_y(_vectors(s1_prime), tolerance)
    s2_vecy = _first_y(_vectors(s2_prime), tolerance)

    if s1_vecy and s2_vecy:
        affine = Affine2D.compose_ltr(
            (
                s1_to_origin,
                s1_vec1_to_s2_vec1,
                # lie vec1 along x axis
                rotate_s2vec1_onto_x,
                # scale first y-vectors to match; x-parts should already match
                Affine2D.identity().scale(1.0, s2_vecy.y / s1_vecy.y),
                # restore the rotation we removed
                rotate_s2vec1_off_x,
                # drop into final position
                origin_to_s2,
            )
        )
        if _try_affine(affine, s1, s2, tolerance):
            return _round(affine, s1, s2, tolerance)

    # If we still aren't the same give up
    return None



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from typing import Generator, Tuple
from picosvg import svg_meta

_CMD_RE = re.compile(f'([{"".join(svg_meta.cmds())}])')
_SEPARATOR_RE = re.compile("[, ]+")
_FLOAT_RE = re.compile(
    r"[-+]?"  # optional sign
    r"(?:"
    r"(?:0|[1-9][0-9]*)(?:\.[0-9]+)?"  # int or float
    r"|"
    r"(?:\.[0-9]+)"  # float with leading dot (e.g. '.42')
    r")"
    r"(?:[eE][-+]?[0-9]+)?"  # optional scientific notiation
)
_BOOL_RE = re.compile("^[01]")
_ARC_ARGUMENT_TYPES = (
    (float, _FLOAT_RE),  # rx
    (float, _FLOAT_RE),  # ry
    (float, _FLOAT_RE),  # x-axis-rotation
    (int, _BOOL_RE),  # large-arc-flag
    (int, _BOOL_RE),  # sweep-flag
    (float, _FLOAT_RE),  # x
    (float, _FLOAT_RE),  # y
)

# https://www.w3.org/TR/SVG11/paths.html#PathDataMovetoCommands
# If a moveto is followed by multiple pairs of coordinates,
# the subsequent pairs are treated as implicit lineto commands
_IMPLICIT_REPEAT_CMD = {"m": "l", "M": "L"}


def _parse_args(cmd: str, args: str) -> Generator[float, None, None]:
    raw_args = [s for s in _SEPARATOR_RE.split(args) if s]
    if not raw_args:
        return

    if cmd.upper() == "A":
        arg_types = _ARC_ARGUMENT_TYPES
    else:
        arg_types = ((float, _FLOAT_RE),)
    n = len(arg_types)

    i = j = 0
    while j < len(raw_args):
        arg = raw_args[j]
        # modulo to wrap around
        converter, regex = arg_types[i % n]
        m = regex.match(arg)
        if not m:
            raise ValueError(f"Invalid argument #{i} for '{cmd}': {arg!r}")

        start, end = m.span()
        yield converter(arg[start:end])

        if end < len(arg):
            raw_args[j] = arg[end:]
        else:
            j += 1
        i += 1


def _explode_cmd(args_per_cmd, cmd, args):
    cmds = []
    for i in range(len(args) // args_per_cmd):
        if i > 0:
            cmd = _IMPLICIT_REPEAT_CMD.get(cmd, cmd)
        cmds.append((cmd, tuple(args[i * args_per_cmd : (i + 1) * args_per_cmd])))
    return cmds


def parse_svg_path(
    svg_path: str, exploded: bool = False
) -> Generator[Tuple[str, Tuple[float, ...]], None, None]:
    """Parses an svg path.

    Exploded means when params repeat each the command is reported as
    if multiplied. For example "M1,1 2,2 3,3" would report as three
    separate steps when exploded.

    Yields tuples of (cmd, (args))."""
    command_tuples = []
    parts = _CMD_RE.split(svg_path)[1:]
    for i in range(0, len(parts), 2):
        cmd = parts[i]
        raw_args = parts[i + 1].strip()

        args = tuple(_parse_args(cmd, raw_args))

        args_per_cmd = svg_meta.check_cmd(cmd, args)
        if args_per_cmd == 0 or not exploded:
            command_tuples.append((cmd, args))
        else:
            command_tuples.extend(_explode_cmd(args_per_cmd, cmd, args))
    for t in command_tuples:
        yield t
